# ==================================================================================================================== #
#              _____ ____    _        _      ___ ______  __    _    ____ _____                                         #
#  _ __  _   _| ____|  _ \  / \      / \    |_ _|  _ \ \/ /   / \  / ___|_   _|                                        #
# | '_ \| | | |  _| | | | |/ _ \    / _ \    | || |_) \  /   / _ \| |     | |                                          #
# | |_) | |_| | |___| |_| / ___ \  / ___ \ _ | ||  __//  \  / ___ \ |___  | |                                          #
# | .__/ \__, |_____|____/_/   \_\/_/   \_(_)___|_|  /_/\_\/_/   \_\____| |_|                                          #
# |_|    |___/                                                                                                         #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2017-2026 Patrick Lehmann - Bötzingen, Germany                                                             #
# Copyright 2016-2016 Patrick Lehmann - Dresden, Germany                                                               #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
from pathlib              import Path
from textwrap             import dedent
from typing               import ClassVar, Optional as Nullable, List

from lxml.etree           import _Element, QName
from pyTooling.Decorators import export

from pyEDAA.IPXACT        import RootElement, __DEFAULT_SCHEMA__, VLNV, IPXACTException


@export
class GeneratorChain(RootElement):
	"""Represents an IP-XACT generator chain."""

	_rootTagName:                  ClassVar[str] = "generatorChain"

	_displayName:                  str
	_chainGroup:                   List
	_generatorChainSelector:       "GeneratorChainSelector"
	_interconnectionConfiguration: List
	_generator:                    "Generator"

	def __init__(
		self,
		generatorChainFile: Nullable[Path] = None,
		parse: bool = False,
		vlnv: Nullable[VLNV] = None,
		description: Nullable[str] = None
	) -> None:
		self._displayName =                   ""  # displayName
		self._chainGroup =                    []  # chainGroup
		self._generatorChainSelector =        None
		self._interconnectionConfiguration =  None
		self._generator =                     None

		super().__init__(generatorChainFile, parse, vlnv, description)

	def Parse(self, element: _Element) -> None:
		elementLocalname = QName(element).localname
		# if elementLocalname == "catalogs":
		# 	for ipxactFileElement in element:
		# 		self.AddItem(IpxactFile.FromXml(ipxactFileElement))
		# else:
		raise IPXACTException(f"Unsupported tag '{elementLocalname}' at root-level.")

	def SetItem(self, item):
		if isinstance(item,   GeneratorChainSelector):      self._generatorChainSelector =      item
		elif isinstance(item, ComponentGeneratorSelector):  self._componentGeneratorSelector =  item
		elif isinstance(item, Generator):                   self._generator =                   item
		else:
			raise ValueError()

	def ToXml(self) -> str:
		"""Converts the object's data into XML format."""

		buffer = dedent("""\
			<?xml version="1.0" encoding="UTF-8"?>
			<{xmlns}:generatorChain
				xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
				xmlns:{xmlns}="{schemaUri}"
				xsi:schemaLocation="{schemaUri} {schemaUrl}">
			{versionedIdentifier}
				<{xmlns}:displayName>{displayName}</{xmlns}:displayName>
				<{xmlns}:description>{description}</{xmlns}:description>
				<{xmlns}:chainGroup>{chainGroup}</{xmlns}:chainGroup>
			""").format(
				xmlns=__DEFAULT_SCHEMA__.NamespacePrefix,
				schemaUri=__DEFAULT_SCHEMA__.SchemaUri,
				schemaUrl=__DEFAULT_SCHEMA__.SchemaUrl,
				versionedIdentifier=self._vlnv.ToXml(isVersionedIdentifier=True),
				displayName=self._displayName,
				description=self._description,
				chainGroup=self._chainGroup
			)

		if self._generatorChainSelector:
			buffer += "\t<{xmlns}:generatorChainSelector>\n"
			buffer += self._generatorChainSelector.ToXml(2)
			buffer += "\t</{xmlns}:generatorChainSelector>\n"

		if self._componentGeneratorSelector:
			buffer += "\t<{xmlns}:componentGeneratorSelector>\n"
			buffer += self._componentGeneratorSelector.ToXml(2)
			buffer += "\t</{xmlns}:componentGeneratorSelector>\n"

		if self._generator:
			buffer += "\t<{xmlns}:generator>\n"
			buffer += self._generator.ToXml(2)
			buffer += "\t</{xmlns}:generator>\n"

		buffer += dedent("""\
			</{xmlns}:generatorChain>
			""")

		return buffer.format(xmlns=__DEFAULT_SCHEMA__.NamespacePrefix)


@export
class GeneratorChainSelector:
	"""Represents an IP-XACT generator chain selector."""

	def __init__(self) -> None:
		pass

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class ComponentGeneratorSelector:
	"""Represents an IP-XACT component generator selector."""

	def __init__(self) -> None:
		pass

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Generator:
	"""Represents an IP-XACT generator."""

	def __init__(self) -> None:
		pass

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""



# ==================================================================================================================== #
#              _____ ____    _        _      ___ ______  __    _    ____ _____                                         #
#  _ __  _   _| ____|  _ \  / \      / \    |_ _|  _ \ \/ /   / \  / ___|_   _|                                        #
# | '_ \| | | |  _| | | | |/ _ \    / _ \    | || |_) \  /   / _ \| |     | |                                          #
# | |_) | |_| | |___| |_| / ___ \  / ___ \ _ | ||  __//  \  / ___ \ |___  | |                                          #
# | .__/ \__, |_____|____/_/   \_\/_/   \_(_)___|_|  /_/\_\/_/   \_\____| |_|                                          #
# |_|    |___/                                                                                                         #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2017-2026 Patrick Lehmann - Bötzingen, Germany                                                             #
# Copyright 2016-2016 Patrick Lehmann - Dresden, Germany                                                               #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
from pathlib              import Path
from sys                  import version_info
from textwrap             import dedent
from typing import List, Optional as Nullable, ClassVar, Dict, Union

from lxml.etree           import _Element, QName, _Comment
from pyTooling.Decorators import export, readonly
from pyTooling.Common     import getFullyQualifiedName

from pyEDAA.IPXACT        import __DEFAULT_SCHEMA__, RootElement, VLNV, IPXACTSchema, Element, IPXACTException


@export
class BusInterface(Element):
	"""Represents an IP-XACT bus interface."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	@classmethod
	def FromXml(cls, element: _Element) -> "BusInterface":
		pass

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class IndirectInterface(Element):
	"""Represents an IP-XACT indirect interface."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Channel(Element):
	"""Represents an IP-XACT channel."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class RemapState(Element):
	"""Represents an IP-XACT remap state."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class AddressSpace(Element):
	"""Represents an IP-XACT address space."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class MemoryMap(Element):
	"""Represents an IP-XACT memory map."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Model(Element):
	"""Represents an IP-XACT model."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class ComponentGenerator:
	"""Represents an IP-XACT component generator."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Choice(Element):
	"""Represents an IP-XACT choice."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class File(Element):
	_path: Path

	def __init__(self, path: Path) -> None:
		self._path = Path

	@readonly
	def Path(self) -> Path:
		return self._path

	@classmethod
	def FromXml(cls, fileElement: _Element) -> "FileSet":
		fileName = None
		fileType = None
		for element in fileElement:
			if isinstance(element, _Comment):
				continue

			elementLocalname = QName(element).localname
			if elementLocalname == "name":
				fileName = element.text
			elif elementLocalname == "fileType":
				fileType = element.text
			elif elementLocalname == "isStructural":
				pass
			else:
				raise IPXACTException(f"Unsupported tag '{elementLocalname}' at component → fileSets → fileSet → file.")

		return cls(Path(fileName))

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""

	def __str__(self) -> str:
		return str(self._path)


@export
class FileSet(Element):
	"""Represents an IP-XACT fileset."""

	_name: str
	_files: List[File]

	def __init__(self, name: str, files: List[File]) -> None:
		self._name =  name
		self._files = [file for file in files]

	@readonly
	def Name(self) -> str:
		return self._name

	@readonly
	def Files(self) -> List[File]:
		return self._files

	@readonly
	def FileCount(self) -> int:
		return len(self._files)

	@classmethod
	def FromXml(cls, fileSetElement: _Element) -> "FileSet":
		files = []
		fileSetName = None
		for element in fileSetElement:
			if isinstance(element, _Comment):
				continue

			elementLocalname = QName(element).localname
			if elementLocalname == "name":
				fileSetName = element.text
			elif elementLocalname == "file":
				files.append(File.FromXml(element))
			else:
				raise IPXACTException(f"Unsupported tag '{elementLocalname}' at component → fileSets → fileSet.")

		return cls(fileSetName, files)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""

	def __str__(self) -> str:
		return f"FileSet {self._name} ({len(self._files)})"


@export
class WhiteboxElement(Element):
	"""Represents an IP-XACT whitebos element."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Cpu(Element):
	"""Represents an IP-XACT cpu."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class OtherClockDriver(Element):
	"""Represents an IP-XACT *other* clock driver."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class ResetType(Element):
	"""Represents an IP-XACT reset type."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Parameter(Element):
	"""Represents an IP-XACT parameter.""""""Represents an IP-XACT assertion."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Assertion(Element):
	"""Represents an IP-XACT assertion."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent=0):
		"""Converts the object's data into XML format."""

		return ""


@export
class Component(RootElement):
	"""Represents an IP-XACT components."""

	_rootTagName:         ClassVar[str] = "component"

	_busInterfaces:       List
	_indirectInterfaces:  List
	_channels:            List
	_remapStates:         List
	_addressSpaces:       List
	_memoryMaps:          List
	_model:               Nullable[Model]
	_componentGenerators: List
	_choices:             List
	_fileSets:            Dict[str, FileSet]
	_whiteboxElements:    List
	_cpus:                List
	_otherClockDrivers:   List
	_resetTypes:          List
	_parameters:          List
	_assertions:          List

	def __init__(
		self,
		componentFile: Nullable[Path] = None,
		parse: bool = False,
		vlnv: Nullable[VLNV] = None,
		description: Nullable[str] = None
	) -> None:
		self._busInterfaces = []
		self._indirectInterfaces = []
		self._channels = []
		self._remapStates = []
		self._addressSpaces = []
		self._memoryMaps = []
		self._model = None
		self._componentGenerators = []
		self._choices = []
		self._fileSets = {}
		self._whiteboxElements = []
		self._cpus = []
		self._otherClockDrivers = []
		self._resetTypes = []
		self._parameters = []
		self._assertions = []

		super().__init__(componentFile, parse, vlnv, description)

	@readonly
	def FileSets(self) -> Dict[str, FileSet]:
		return self._fileSets

	def Parse(self, element: _Element) -> None:
		elementLocalname = QName(element).localname
		if elementLocalname == "busInterfaces":
			pass
		# for busInterfaceElement in element:
		# 	self.AddItem(BusInterface.FromXml(busInterfaceElement))
		elif elementLocalname == "indirectInterfaces":
			pass
		elif elementLocalname == "channels":
			pass
		elif elementLocalname == "remapStates":
			pass
		elif elementLocalname == "addressSpaces":
			pass
		elif elementLocalname == "memoryMaps":
			pass
		elif elementLocalname == "model":
			pass
		elif elementLocalname == "componentGenerators":
			pass
		elif elementLocalname == "choices":
			pass
		elif elementLocalname == "fileSets":
			for fileSetElement in element:
				if isinstance(fileSetElement, _Comment):
					continue

				self.AddFileSet(FileSet.FromXml(fileSetElement))
		elif elementLocalname == "whiteboxElements":
			pass
		elif elementLocalname == "cpus":
			pass
		elif elementLocalname == "otherClockDrivers":
			pass
		elif elementLocalname == "resetTypes":
			pass
		elif elementLocalname == "parameters":
			pass
		elif elementLocalname == "assertions":
			pass
		else:
			raise IPXACTException(f"Unsupported tag '{elementLocalname}' at root-level.")

	def SetItem(self, item):
		if isinstance(item, Model):
			self._model = item
		else:
			raise ValueError()

	def AddItem(self, item) -> None:
		if isinstance(item, BusInterface):
			self._busInterfaces.append(item)
		elif isinstance(item, IndirectInterface):
			self._indirectInterfaces.append(item)
		elif isinstance(item, Channel):
			self._channels.append(item)
		elif isinstance(item, RemapState):
			self._remapStates.append(item)
		elif isinstance(item, AddressSpace):
			self._addressSpaces.append(item)
		elif isinstance(item, MemoryMap):
			self._memoryMaps.append(item)
		elif isinstance(item, ComponentGenerator):
			self._componentGenerators.append(item)
		elif isinstance(item, Choice):
			self._choices.append(item)
		elif isinstance(item, FileSet):
			self.AddFileSet(item)
		elif isinstance(item, WhiteboxElement):
			self._whiteboxElements.append(item)
		elif isinstance(item, Cpu):
			self._cpus.append(item)
		elif isinstance(item, OtherClockDriver):
			self._otherClockDrivers.append(item)
		elif isinstance(item, ResetType):
			self._resetTypes.append(item)
		elif isinstance(item, Parameter):
			self._parameters.append(item)
		elif isinstance(item, Assertion):
			self._assertions.append(item)
		else:
			ex = TypeError("Parameter 'item' is not a BusInterface, IndirectInterface, Channel, RemapState, AddressSpace, MemoryMap, ComponentGenerator, Choice, FileSet, WhiteboxElement, Cpu, OtherClockDriver, ResetType, Parameter, or Assertion.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(item)}'.")
			raise ex

	def AddFileSet(self, fileset: FileSet):
		if not isinstance(fileset, FileSet):
			ex = TypeError("Parameter 'fileset' is not a FileSet.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(fileset)}'.")
			raise ex

		if fileset._name in self._fileSets:
			raise ValueError(f"Duplicate fileset '{fileset._name}'.")

		self._fileSets[fileset._name] = fileset

	def ToXml(self, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		xmlns = schema.NamespacePrefix
		buffer = dedent(f"""\
			<?xml version="1.0" encoding="UTF-8" ?>
			<{xmlns}:component
			\txmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
			\txmlns:{xmlns}="{schema.SchemaUri}"
			\txsi:schemaLocation="{schema.SchemaUri} {schema.SchemaUrl}">
			{self._vlnv.ToXml(schema, isVersionedIdentifier=True)}
			\t<{xmlns}:description>{self._description}</{xmlns}:description>
			""")

		if self._busInterfaces:
			buffer += f"\t<{xmlns}:busInterfaces>\n"
			for busInterface in self._busInterfaces:
				buffer += busInterface.ToXml(2, schema)
			buffer += f"\t</{xmlns}:busInterfaces>\n"

		if self._indirectInterfaces:
			buffer += f"\t<{xmlns}:indirectInterfaces>\n"
			for indirectInterface in self._indirectInterfaces:
				buffer += indirectInterface.ToXml(2, schema)
			buffer += f"\t</{xmlns}:indirectInterfaces>\n"

		if self._channels:
			buffer += f"\t<{xmlns}:channels>\n"
			for channel in self._channels:
				buffer += channel.ToXml(2, schema)
			buffer += f"\t</{xmlns}:channels>\n"

		if self._remapStates:
			buffer += f"\t<{xmlns}:remapStates>\n"
			for remapState in self._remapStates:
				buffer += remapState.ToXml(2, schema)
			buffer += f"\t</{xmlns}:remapStates>\n"

		if self._addressSpaces:
			buffer += f"\t<{xmlns}:addressSpaces>\n"
			for addressSpace in self._addressSpaces:
				buffer += addressSpace.ToXml(2, schema)
			buffer += f"\t</{xmlns}:addressSpaces>\n"

		if self._memoryMaps:
			buffer += f"\t<{xmlns}:memoryMaps>\n"
			for memoryMap in self._memoryMaps:
				buffer += memoryMap.ToXml(2, schema)
			buffer += f"\t</{xmlns}:memoryMaps>\n"

		if self._model:
			buffer += f"\t<{xmlns}:model>\n"
			buffer += self._model.ToXml(2, schema)
			buffer += f"\t</{xmlns}:model>\n"

		if self._componentGenerators:
			buffer += f"\t<{xmlns}:componentGenerators>\n"
			for componentGenerator in self._componentGenerators:
				buffer += componentGenerator.ToXml(2, schema)
			buffer += f"\t</{xmlns}:componentGenerators>\n"

		if self._choices:
			buffer += f"\t<{xmlns}:choices>\n"
			for choice in self._choices:
				buffer += choice.ToXml(2, schema)
			buffer += f"\t</{xmlns}:choices>\n"

		if self._fileSets:
			buffer += f"\t<{xmlns}:fileSets>\n"
			for fileSet in self._fileSets:
				buffer += fileSet.ToXml(2, schema)
			buffer += f"\t</{xmlns}:fileSets>\n"

		if self._whiteboxElements:
			buffer += f"\t<{xmlns}:whiteboxElements>\n"
			for whiteboxElement in self._whiteboxElements:
				buffer += whiteboxElement.ToXml(2, schema)
			buffer += f"\t</{xmlns}:whiteboxElements>\n"

		if self._cpus:
			buffer += f"\t<{xmlns}:cpus>\n"
			for cpu in self._cpus:
				buffer += cpu.ToXml(2, schema)
			buffer += f"\t</{xmlns}:cpus>\n"

		if self._otherClockDrivers:
			buffer += f"\t<{xmlns}:otherClockDrivers>\n"
			for otherClockDriver in self._otherClockDrivers:
				buffer += otherClockDriver.ToXml(2, schema)
			buffer += f"\t</{xmlns}:otherClockDrivers>\n"

		if self._resetTypes:
			buffer += f"\t<{xmlns}:resetTypes>\n"
			for resetType in self._resetTypes:
				buffer += resetType.ToXml(2, schema)
			buffer += f"\t</{xmlns}:resetTypes>\n"

		if self._parameters:
			buffer += f"\t<{xmlns}:parameters>\n"
			for parameter in self._parameters:
				buffer += parameter.ToXml(2, schema)
			buffer += f"\t</{xmlns}:parameters>\n"

		if self._assertions:
			buffer += f"\t<{xmlns}:assertions>\n"
			for assertion in self._assertions:
				buffer += assertion.ToXml(2, schema)
			buffer += f"\t</{xmlns}:assertions>\n"

		buffer += dedent(f"""\
				</{xmlns}:component>
				""")

		return buffer



# ==================================================================================================================== #
#              _____ ____    _        _      ___ ______  __    _    ____ _____                                         #
#  _ __  _   _| ____|  _ \  / \      / \    |_ _|  _ \ \/ /   / \  / ___|_   _|                                        #
# | '_ \| | | |  _| | | | |/ _ \    / _ \    | || |_) \  /   / _ \| |     | |                                          #
# | |_) | |_| | |___| |_| / ___ \  / ___ \ _ | ||  __//  \  / ___ \ |___  | |                                          #
# | .__/ \__, |_____|____/_/   \_\/_/   \_(_)___|_|  /_/\_\/_/   \_\____| |_|                                          #
# |_|    |___/                                                                                                         #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2017-2026 Patrick Lehmann - Bötzingen, Germany                                                             #
# Copyright 2016-2016 Patrick Lehmann - Dresden, Germany                                                               #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
from pathlib              import Path
from textwrap             import dedent
from typing               import Optional as Nullable, ClassVar

from lxml.etree           import _Element, QName
from pyTooling.Decorators import export

from pyEDAA.IPXACT        import RootElement, __DEFAULT_SCHEMA__, VLNV, IPXACTSchema, IPXACTException


@export
class DesignConfiguration(RootElement):
	"""Represents an IP-XACT design configuration."""

	_rootTagName:                  ClassVar[str] = "designConfiguration"

	_generatorChainConfiguration:  Nullable["GeneratorChainConfiguration"]
	_interconnectionConfiguration: Nullable["InterconnectionConfiguration"]
	_viewConfiguration:            Nullable["ViewConfiguration"]

	def __init__(
		self,
		designConfigurationFile: Nullable[Path] = None,
		parse: bool = False,
		vlnv: Nullable[VLNV] = None,
		description: Nullable[str] = None
	) -> None:
		self._generatorChainConfiguration =  None
		self._interconnectionConfiguration = None
		self._viewConfiguration =            None

		super().__init__(designConfigurationFile, parse, vlnv, description)

	def Parse(self, element: _Element) -> None:
		elementLocalname = QName(element).localname
		# if elementLocalname == "catalogs":
		# 	for ipxactFileElement in element:
		# 		self.AddItem(IpxactFile.FromXml(ipxactFileElement))
		# else:
		raise IPXACTException(f"Unsupported tag '{elementLocalname}' at root-level.")

	def SetItem(self, item):
		if isinstance(item,   GeneratorChainConfiguration):
			self._generatorChainConfiguration =   item
		elif isinstance(item, InterconnectionConfiguration):
			self._interconnectionConfiguration =  item
		elif isinstance(item, ViewConfiguration):
			self._viewConfiguration =             item
		else:
			raise ValueError()

	def ToXml(self, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		xmlns = schema.NamespacePrefix
		buffer = dedent(f"""\
			<?xml version="1.0" encoding="UTF-8"?>
			<{xmlns}:designConfiguration
			\txmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
			\txmlns:{xmlns}="{schema.SchemaUri}"
			\txsi:schemaLocation="{schema.SchemaUri} {schema.SchemaUrl}">
			{self._vlnv.ToXml(schema, isVersionedIdentifier=True)}
			\t<{xmlns}:description>{self._description}</{xmlns}:description>
			""")

		if self._generatorChainConfiguration:
			buffer += f"\t<{xmlns}:componentInstances>\n"
			buffer += self._generatorChainConfiguration.ToXml(2, schema)
			buffer += f"\t</{xmlns}:componentInstances>\n"

		if self._interconnectionConfiguration:
			buffer += f"\t<{xmlns}:interconnectionConfiguration>\n"
			buffer += self._interconnectionConfiguration.ToXml(2, schema)
			buffer += f"\t</{xmlns}:interconnectionConfiguration>\n"

		if self._viewConfiguration:
			buffer += f"\t<{xmlns}:viewConfiguration>\n"
			buffer += self._viewConfiguration.ToXml(2, schema)
			buffer += f"\t</{xmlns}:viewConfiguration>\n"

		buffer += dedent(f"""\
			</{xmlns}:designConfiguration>
			""")

		return buffer


@export
class GeneratorChainConfiguration:
	"""Represents an IP-XACT generator chain configuration."""

	def __init__(self) -> None:
		pass

	def ToXml(self, indent: int = 0, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		return ""


@export
class InterconnectionConfiguration:
	"""Represents an IP-XACT interconnection configuration."""

	def __init__(self) -> None:
		pass

	def ToXml(self, indent: int = 0, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		return ""


@export
class ViewConfiguration:
	"""Represents an IP-XACT view configuration."""

	def __init__(self) -> None:
		pass

	def ToXml(self, indent: int = 0, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		return ""



# ==================================================================================================================== #
#               ___ ______  __    _    ____ _____                                                                      #
#   _ __  _   _|_ _|  _ \ \/ /   / \  / ___|_   _|                                                                     #
#  | '_ \| | | || || |_) \  /   / _ \| |     | |                                                                       #
#  | |_) | |_| || ||  __//  \  / ___ \ |___  | |                                                                       #
#  | .__/ \__, |___|_|  /_/\_\/_/   \_\____| |_|                                                                       #
#  |_|    |___/                                                                                                        #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2017-2025 Patrick Lehmann - Bötzingen, Germany                                                             #
# Copyright 2016-2016 Patrick Lehmann - Dresden, Germany                                                               #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
"""
This package is a resource package containing various data files.

.. rubric:: Common Files

* :file:`README.md`
* :file:`LICENSE.md`
* :file:`NOTICE.md`

.. rubric:: XML Schema Files

* :file:`ipxact-1.0/`

  * :file:`README.md`
  * :file:`index.xsd`
  * :file:`***.xsd`

* :file:`ipxact-1.1/`

  * :file:`README.md`
  * :file:`index.xsd`
  * :file:`***.xsd`

* :file:`ipxact-1.2/`

  * :file:`README.md`
  * :file:`index.xsd`
  * :file:`***.xsd`

* :file:`ipxact-1.4/`

  * :file:`README.md`
  * :file:`index.xsd`
  * :file:`***.xsd`

* :file:`ipxact-1.5/`

  * :file:`README.md`
  * :file:`index.xsd`
  * :file:`***.xsd`

* :file:`ieee-1685-2009/`

  * :file:`README.md`
  * :file:`index.xsd`
  * :file:`***.xsd`

* :file:`ieee-1685-2014/`

  * :file:`README.md`
  * :file:`index.xsd`
  * :file:`***.xsd`

* :file:`ieee-1685-2022/`

  * :file:`README.md`
  * :file:`LICENSE`
  * :file:`NOTICE`
  * :file:`index.xsd`
  * :file:`***.xsd`

"""
from sys import modules

from pyTooling.Common import getResourceFile

__author__ =    "Patrick Lehmann"
__email__ =     "Paebbels@gmail.com"
__copyright__ = "2016-2025, Patrick Lehmann"
__license__ =   "Apache License, Version 2.0"
__version__ =   "0.4.0"

__module = modules[__name__]

_IPXACT_10_INDEX = getResourceFile(__module, "ipxact-1.0/index.xsd")
_IPXACT_11_INDEX = getResourceFile(__module, "ipxact-1.1/index.xsd")
_IPXACT_12_INDEX = getResourceFile(__module, "ipxact-1.2/index.xsd")
_IPXACT_14_INDEX = getResourceFile(__module, "ipxact-1.4/index.xsd")
_IPXACT_15_INDEX = getResourceFile(__module, "ipxact-1.5/index.xsd")

_IPXACT_2009_INDEX = getResourceFile(__module, "ieee-1685-2009/index.xsd")
_IPXACT_2014_INDEX = getResourceFile(__module, "ieee-1685-2014/index.xsd")
_IPXACT_2022_INDEX = getResourceFile(__module, "ieee-1685-2022/index.xsd")

__all__ = [
	"_IPXACT_10_INDEX",
	"_IPXACT_11_INDEX",
	"_IPXACT_12_INDEX",
	"_IPXACT_14_INDEX",
	"_IPXACT_15_INDEX",
	"_IPXACT_2009_INDEX",
	"_IPXACT_2014_INDEX",
	"_IPXACT_2022_INDEX"
]



# ==================================================================================================================== #
#              _____ ____    _        _      ___ ______  __    _    ____ _____                                         #
#  _ __  _   _| ____|  _ \  / \      / \    |_ _|  _ \ \/ /   / \  / ___|_   _|                                        #
# | '_ \| | | |  _| | | | |/ _ \    / _ \    | || |_) \  /   / _ \| |     | |                                          #
# | |_) | |_| | |___| |_| / ___ \  / ___ \ _ | ||  __//  \  / ___ \ |___  | |                                          #
# | .__/ \__, |_____|____/_/   \_\/_/   \_(_)___|_|  /_/\_\/_/   \_\____| |_|                                          #
# |_|    |___/                                                                                                         #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2017-2026 Patrick Lehmann - Bötzingen, Germany                                                             #
# Copyright 2016-2016 Patrick Lehmann - Dresden, Germany                                                               #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
from pathlib              import Path
from sys                  import version_info
from textwrap             import dedent
from typing               import List, ClassVar, Optional as Nullable

from lxml.etree           import _Element, QName
from pyTooling.Decorators import export
from pyTooling.Common     import getFullyQualifiedName

from pyEDAA.IPXACT        import RootElement, __DEFAULT_SCHEMA__, VLNV, IPXACTSchema, Element, IPXACTException


@export
class Design(RootElement):
	"""Represents an IP-XACT design."""

	_rootTagName:            ClassVar[str] = "design"

	_componentInstances: List
	_interconnections:   List
	_adHocConnections:   List

	def __init__(
		self,
		designFile: Nullable[Path] = None,
		parse: bool = False,
		vlnv: Nullable[VLNV] = None,
		description: Nullable[str] = None
	) -> None:
		"""
		Instantiates a design structure.

		:param vlnv:        A Vendor-Library-Name-Version unique identified.
		:param description: A description text.
		"""
		self._componentInstances = []
		self._interconnections =   []
		self._adHocConnections =   []

		super().__init__(designFile, parse, vlnv, description)

		# if not isinstance(description, str):
		# 	ex = TypeError(f"Parameter 'description' is not a string.")
		# 	if version_info >= (3, 11):  # pragma: no cover
		# 		ex.add_note(f"Got type '{getFullyQualifiedName(description)}'.")
		# 	raise ex
		# elif description == "":
		# 	raise ValueError(f"Parameter 'description' is empty.")

	def Parse(self, element: _Element) -> None:
		elementLocalname = QName(element).localname
		if elementLocalname == "componentInstances":
			pass
			# for ipxactFileElement in element:
			# 	self.AddItem(IpxactFile.FromXml(ipxactFileElement))
		elif elementLocalname == "componentInstances":
			pass
		elif elementLocalname == "interconnections":
			pass
		elif elementLocalname == "adHocConnections":
			pass
		else:
			raise IPXACTException(f"Unsupported tag '{elementLocalname}' at root-level.")

	def AddItem(self, item) -> None:
		if isinstance(item, ComponentInstance):
			self._componentInstances.append(item)
		elif isinstance(item, Interconnection):
			self._interconnections.append(item)
		elif isinstance(item, AdHocConnection):
			self._adHocConnections.append(item)
		else:
			raise ValueError()

	def ToXml(self, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		xmlns = schema.NamespacePrefix
		buffer = dedent(f"""\
			<?xml version="1.0" encoding="UTF-8"?>
			<{xmlns}:design
			\txmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
			\txmlns:{xmlns}="{schema.SchemaUri}"
			\txsi:schemaLocation="{schema.SchemaUri} {schema.SchemaUrl}">
			{self._vlnv.ToXml(schema, isVersionedIdentifier=True)}
			\t<{xmlns}:description>{self._description}</{xmlns}:description>
			""")

		if self._componentInstances:
			buffer += f"\t<{xmlns}:componentInstances>\n"
			for componentInstance in self._componentInstances:
				buffer += componentInstance.ToXml(2, schema)
			buffer += f"\t</{xmlns}:componentInstances>\n"

		if self._interconnections:
			buffer += f"\t<{xmlns}:interconnections>\n"
			for interconnection in self._interconnections:
				buffer += interconnection.ToXml(2, schema)
			buffer += f"\t</{xmlns}:interconnections>\n"

		if self._adHocConnections:
			buffer += f"\t<{xmlns}:adHocConnections>\n"
			for adHocConnection in self._adHocConnections:
				buffer += adHocConnection.ToXml(2, schema)
			buffer += f"\t</{xmlns}:adHocConnections>\n"

		buffer += dedent(f"""\
			</{xmlns}:design>
			""")

		return buffer


@export
class IpxactFile(Element):
	"""Represents a IP-XACT file."""

	_name:        str    #: Name
	_description: str    #: Description

	def __init__(self, vlnv, name, description) -> None:
		"""
		Instantiates an ipxactFile structure.

		:param vlnv:        A Vendor-Library-Name-Version unique identified.
		:param name:        Name of the IP-XACT file.
		:param description: A description text.
		:raises TypeError:  If parameter vlnv is not a VLNV.
		:raises TypeError:  If parameter name is not a string.
		:raises ValueError: If parameter name is empty.
		:raises TypeError:  If parameter description is not a string.
		:raises ValueError: If parameter description is empty.
		"""
		super().__init__(vlnv)

		if not isinstance(name, str):
			ex = TypeError(f"Parameter 'name' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(name)}'.")
			raise ex
		elif name == "":
			raise ValueError(f"Parameter 'name' is empty.")

		if not isinstance(description, str):
			ex = TypeError(f"Parameter 'description' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(description)}'.")
			raise ex
		elif description == "":
			raise ValueError(f"Parameter 'description' is empty.")

		self._name = name
		self._description = description

	def ToXml(self, indent: int = 0, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		# WORKAROUND:
		#   Python <=3.11:
		#   {'\t' * indent} is not supported by Python before 3.12 due to a backslash within {...}
		indent = "\t" * indent
		xmlns = schema.NamespacePrefix
		return dedent(f"""\
			{indent}<{xmlns}:ipxactFile>
			{indent}\t{self._vlnv.ToXml(indent)}
			{indent}\t<{xmlns}:name>{self._name}</{xmlns}:name>
			{indent}\t<{xmlns}:description>{self._description}</{xmlns}:description>
			{indent}</{xmlns}:ipxactFile>
		""")


@export
class ComponentInstance(Element):
	"""Represents an IP-XACT component instance."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent: int = 0, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		return ""


@export
class Interconnection(Element):
	"""Represents an IP-XACT interconnection."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent: int = 0, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		return ""


@export
class AdHocConnection(Element):
	"""Represents an IP-XACT ad-hoc connection."""

	def __init__(self, vlnv: VLNV) -> None:
		super().__init__(vlnv)

	def ToXml(self, indent: int = 0, schema: IPXACTSchema = __DEFAULT_SCHEMA__) -> str:
		"""Converts the object's data into XML format."""

		return ""



# ==================================================================================================================== #
#              _____ ____    _        _      ___ ______  __    _    ____ _____                                         #
#  _ __  _   _| ____|  _ \  / \      / \    |_ _|  _ \ \/ /   / \  / ___|_   _|                                        #
# | '_ \| | | |  _| | | | |/ _ \    / _ \    | || |_) \  /   / _ \| |     | |                                          #
# | |_) | |_| | |___| |_| / ___ \  / ___ \ _ | ||  __//  \  / ___ \ |___  | |                                          #
# | .__/ \__, |_____|____/_/   \_\/_/   \_(_)___|_|  /_/\_\/_/   \_\____| |_|                                          #
# |_|    |___/                                                                                                         #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2017-2026 Patrick Lehmann - Bötzingen, Germany                                                             #
# Copyright 2016-2016 Patrick Lehmann - Dresden, Germany                                                               #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
"""A DOM based IP-XACT implementation for Python."""
from pathlib  import Path
from sys      import version_info
from textwrap import dedent
from typing   import Union, Dict, Tuple, Optional as Nullable, ClassVar

from lxml.etree            import XMLParser, XML, XMLSchema, ElementTree, QName, _Element, _Comment
from pyTooling.Decorators  import export, readonly
from pyTooling.MetaClasses import ExtendedType, abstractmethod
from pyTooling.Common      import getFullyQualifiedName
from pyTooling.Versioning  import SemanticVersion, CalendarVersion

from .       import Schema
from .Schema import *

__author__ =    "Patrick Lehmann"
__email__ =     "Paebbels@gmail.com"
__copyright__ = "2016-2026, Patrick Lehmann"
__license__ =   "Apache License, Version 2.0"
__version__ =   "0.6.5"


@export
class IPXACTException(Exception):
	"""Base-exception for all exceptions in this package."""


@export
class IPXACTSchema(metaclass=ExtendedType, slots=True):
	"""Schema descriptor made of version, namespace prefix, URI, URL and local path."""

	_version:         Union[SemanticVersion, CalendarVersion]  #: Schema version
	_namespacePrefix: str                                      #: XML namespace prefix
	_schemaUri:       str                                      #: Schema URI
	_schemaUrl:       str                                      #: Schema URL
	_localPath:       Path                                     #: Local path

	def __init__(
		self,
		version: Union[str, SemanticVersion, CalendarVersion],
		xmlNamespacePrefix: str,
		schemaUri: str,
		schemaUrl: str,
		localPath: Path
	) -> None:
		"""
		Initializes an IP-XACT Schema description.

		:param version:            Version of the IP-XACT Schema.
		:param xmlNamespacePrefix: XML namespace prefix (``<prefix:element>``)
		:param schemaUri:          IP-XACT schema URI
		:param schemaUrl:          URL the IP-XACT schema definition file (XSD).
		:param localPath:          Path to the local XSD file.
		"""
		# TODO: add raises ... lines
		if version is None:
			raise ValueError(f"Parameter 'version' is None.")
		elif isinstance(version, str):
			if version.startswith("20"):
				self._version = CalendarVersion.Parse(version)
			else:
				self._version = SemanticVersion.Parse(version)
		elif isinstance(version, (SemanticVersion, CalendarVersion)):
			self._version = version
		else:
			ex = TypeError(f"Parameter 'version' is neither a 'SemanticVersion', a 'CalendarVersion' nor a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(version)}'.")
			raise ex

		if xmlNamespacePrefix is None:
			raise ValueError(f"Parameter 'namespacePrefix' is None.")
		elif not isinstance(xmlNamespacePrefix, str):
			ex = TypeError(f"Parameter 'namespacePrefix' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(version)}'.")
			raise ex

		if schemaUri is None:
			raise ValueError(f"Parameter 'schemaUri' is None.")
		elif not isinstance(schemaUri, str):
			ex = TypeError(f"Parameter 'schemaUri' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(schemaUri)}'.")
			raise ex

		if schemaUrl is None:
			raise ValueError(f"Parameter 'schemaUrl' is None.")
		elif not isinstance(schemaUrl, str):
			ex = TypeError(f"Parameter 'schemaUrl' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(schemaUrl)}'.")
			raise ex

		if localPath is None:
			raise ValueError(f"Parameter 'localPath' is None.")
		elif not isinstance(localPath, Path):
			ex = TypeError(f"Parameter 'localPath' is not a Path.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(localPath)}'.")
			raise ex

		self._namespacePrefix = xmlNamespacePrefix
		self._schemaUri =       schemaUri
		self._schemaUrl =       schemaUrl
		self._localPath =       localPath

	@readonly
	def Version(self) -> Union[SemanticVersion, CalendarVersion]:
		return self._version

	@readonly
	def NamespacePrefix(self) -> str:
		return self._namespacePrefix

	@readonly
	def SchemaUri(self) -> str:
		return self._schemaUri

	@readonly
	def SchemaUrl(self) -> str:
		return self._schemaUrl

	@readonly
	def LocalPath(self) -> Path:
		return self._localPath

	def __repr__(self) -> str:
		return f"<{self.__class__.__name__} IP-XACT {self._version} {self._schemaUri} - {self._localPath}>"

	def __str__(self) -> str:
		return f"IP-XACT {self._version}"


#                           version, xmlns,    URI                                                          URL,                                                              Local Path
_IPXACT_10 =   IPXACTSchema("1.0", "spirit", "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1.0", "", _IPXACT_10_INDEX)
_IPXACT_11 =   IPXACTSchema("1.1", "spirit", "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1.1", "", _IPXACT_11_INDEX)
_IPXACT_12 =   IPXACTSchema("1.2", "spirit", "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1.2", "", _IPXACT_12_INDEX)
_IPXACT_14 =   IPXACTSchema("1.4", "spirit", "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1.4", "", _IPXACT_14_INDEX)
_IPXACT_15 =   IPXACTSchema("1.5", "spirit", "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1.5", "", _IPXACT_15_INDEX)
_IPXACT_2009 = IPXACTSchema("2009", "spirit", "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1685-2009", "", _IPXACT_2009_INDEX)
_IPXACT_2014 = IPXACTSchema("2014", "ipxact", "http://www.accellera.org/XMLSchema/IPXACT/1685-2014", "http://www.accellera.org/XMLSchema/IPXACT/1685-2014/index.xsd", _IPXACT_2014_INDEX)
_IPXACT_2022 = IPXACTSchema("2022", "ipxact", "http://www.accellera.org/XMLSchema/IPXACT/1685-2022", "http://www.accellera.org/XMLSchema/IPXACT/1685-2022/index.xsd", _IPXACT_2022_INDEX)

__VERSION_TABLE__: Dict[str, IPXACTSchema] = {
	'1.0':   _IPXACT_10,
	'1.1':   _IPXACT_11,
	'1.4':   _IPXACT_14,
	'1.5':   _IPXACT_15,
	'2009':  _IPXACT_2009,
	'2014':  _IPXACT_2014,
	'2022':  _IPXACT_2022
}          #: Dictionary of all IP-XACT versions mapping to :class:`IpxactSchema` instances.

__URI_MAP__: Dict[str, IPXACTSchema] = {value.SchemaUri: value for key, value in __VERSION_TABLE__.items()}  #: Mapping from schema URIs to :class:`IpxactSchema` instances.

__DEFAULT_VERSION__ = "2022"                                  #: IP-XACT default version
__DEFAULT_SCHEMA__ =  __VERSION_TABLE__[__DEFAULT_VERSION__]  #: IP-XACT default Schema


@export
class VLNV(metaclass=ExtendedType, slots=True):
	"""VLNV data structure (Vendor, Library, Name, Version) as a unique identifier in IP-XACT."""

	_vendor:  str              #: Vendor name in a VLNV unique identifier
	_library: str              #: Library name in a VLNV unique identifier
	_name:    str              #: Component name in a VLNV unique identifier
	_version: SemanticVersion  #: Version in a VLNV unique identifier

	def __init__(self, vendor: str, library: str, name: str, version: Union[str, SemanticVersion]) -> None:
		"""
		Initializes the VLNV data structure.

		:param vendor:  Vendor name in a VLNV unique identifier
		:param library: Library name in a VLNV unique identifier
		:param name:    Component name in a VLNV unique identifier
		:param version: Version in a VLNV unique identifier
		"""

		if vendor is None:
			raise ValueError(f"Parameter 'vendor' is None.")
		elif not isinstance(vendor, str):
			ex = TypeError(f"Parameter 'vendor' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(vendor)}'.")
			raise ex

		if library is None:
			raise ValueError(f"Parameter 'library' is None.")
		elif not isinstance(library, str):
			ex = TypeError(f"Parameter 'library' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(library)}'.")
			raise ex

		if name is None:
			raise ValueError(f"Parameter 'name' is None.")
		elif not isinstance(name, str):
			ex = TypeError(f"Parameter 'name' is not a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(name)}'.")
			raise ex

		if version is None:
			raise ValueError(f"Parameter 'version' is None.")
		elif isinstance(version, str):
			self._version = SemanticVersion.Parse(version)
		elif isinstance(version, SemanticVersion):
			self._version = version
		else:
			ex = TypeError(f"Parameter 'version' is neither a 'SemanticVersion' nor a string.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(version)}'.")
			raise ex

		self._vendor =   vendor
		self._library =  library
		self._name =     name

	@readonly
	def Vendor(self) -> str:
		return self._vendor

	@readonly
	def Library(self) -> str:
		return self._library

	@readonly
	def Name(self) -> str:
		return self._name

	@readonly
	def Version(self) -> SemanticVersion:
		return self._version

	def ToXml(self, indent=1, schema: IPXACTSchema = __DEFAULT_SCHEMA__, isVersionedIdentifier=False) -> str:
		"""
		Converts the object's data into XML format.

		:param indent:                Level of indentations.
		:param schema:                XML schema.
		:param isVersionedIdentifier: If true, generate 4 individual tags (``<vendor>``, ``<library>``, ``<name>``,
		                              ``<version>``), otherwise a single ``<vlnv>``-tag with attributes.
		:returns:                     XML formatted string representation.
		"""

		# WORKAROUND:
		#   Python <=3.11:
		#   {'\t' * indent} is not supported by Python before 3.12 due to a backslash within {...}
		indent = "\t" * indent
		xmlns = schema.NamespacePrefix

		if isVersionedIdentifier:
			return dedent(f"""\
				{indent}<{xmlns}:vendor>{self._vendor}</{xmlns}:vendor>
				{indent}<{xmlns}:library>{self._library}</{xmlns}:library>
				{indent}<{xmlns}:name>{self._name}</{xmlns}:name>
				{indent}<{xmlns}:version>{self._version}</{xmlns}:version>
			""")
		else:
			return f"""{indent}<{xmlns}:vlnv vendor="{self._vendor}" library="{self._library}" name="{self._name}" version="{self._version}"/>"""


@export
class Element(metaclass=ExtendedType, slots=True):
	"""Base-class for all IP-XACT elements."""

	def __init__(self, vlnv: VLNV) -> None:
		"""
		Initializes the Element class.
		"""


@export
class NamedElement(Element):
	"""Base-class for all IP-XACT elements with a VLNV."""

	_vlnv: VLNV   #: VLNV unique identifier.

	def __init__(self, vlnv: VLNV) -> None:
		"""
		Initializes the NameElement with an VLNV field for all derives classes.

		:param vlnv:       VLNV unique identifier.
		:raises TypeError: If parameter vlnv is not a VLNV.
		"""
		if not isinstance(vlnv, VLNV):
			ex = TypeError(f"Parameter 'vlnv' is not a VLNV.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(vlnv)}'.")
			raise ex

		self._vlnv =    vlnv

	@readonly
	def VLNV(self) -> VLNV:
		return self._vlnv


@export
class RootElement(NamedElement):
	"""Base-class for all IP-XACT root elements."""

	_file:        Nullable[Path]
	_rootTagName: ClassVar[str] = ""
	_xmlRoot:     Nullable[_Element]
	_xmlSchema:   Nullable[_Element]

	_description: str

	def __init__(self, file: Nullable[Path] = None, parse: bool = False, vlnv: Nullable[VLNV] = None, description: Nullable[str] = None) -> None:
		self._description = description

		if file is None:
			super().__init__(vlnv)
			self._file = None
		elif isinstance(file, Path):
			self._file = file
			vlnv = None
			if parse:
				self.OpenAndValidate()
				vlnv, self._description = self.ParseVLNVAndDescription()

			super().__init__(vlnv)
		else:
			ex = TypeError(f"Parameter 'file' is not a Path.")
			if version_info >= (3, 11):  # pragma: no cover
				ex.add_note(f"Got type '{getFullyQualifiedName(file)}'.")
			raise ex

	def OpenAndValidate(self) -> None:
		if not self._file.exists():
			raise IPXACTException(f"IPXACT file '{self._file}' not found.") from FileNotFoundError(str(self._file))

		try:
			with self._file.open("rb") as fileHandle:
				content = fileHandle.read()
		except OSError as ex:
			raise IPXACTException(f"Couldn't open '{self._file}'.") from ex

		xmlParser =     XMLParser(remove_blank_text=True, encoding="utf-8")
		self._xmlRoot = XML(content, parser=xmlParser, base_url=self._file.resolve().as_uri())  # - relative paths are not supported
		rootTag =       QName(self._xmlRoot.tag)

		if rootTag.localname != self._rootTagName:
			raise IPXACTException(f"The input IP-XACT file is not a {self._rootTagName} file.")

		namespacePrefix = self._xmlRoot.prefix
		namespaceURI =    self._xmlRoot.nsmap[namespacePrefix]
		if namespaceURI in __URI_MAP__:
			ipxactSchema = __URI_MAP__[namespaceURI]
		else:
			raise IPXACTException(f"The input IP-XACT file uses an unsupported namespace: '{namespaceURI}'.")

		try:
			with ipxactSchema.LocalPath.open("rb") as fileHandle:
				schema = fileHandle.read()
		except OSError as ex:
			raise IPXACTException(f"Couldn't open IP-XACT schema '{ipxactSchema.LocalPath}' for {namespacePrefix} ({namespaceURI}).") from ex

		schemaRoot =      XML(schema, parser=xmlParser, base_url=ipxactSchema.LocalPath.as_uri())
		schemaTree =      ElementTree(schemaRoot)
		self._xmlSchema = XMLSchema(schemaTree)

		try:
			self._xmlSchema.assertValid(self._xmlRoot)
		except Exception as ex:
			raise IPXACTException(f"The input IP-XACT file is not valid according to XML schema {namespaceURI}.") from ex

	def ParseVLNVAndDescription(self) -> Tuple[VLNV, str]:
		vendor = None
		library = None
		name = None
		version = None
		description = None

		found = 0
		i = iter(self._xmlRoot)
		for element in i:
			if isinstance(element, _Comment):
				continue

			elementLocalname = QName(element).localname
			if elementLocalname == "vendor":
				found |= 1
				vendor = element.text
			elif elementLocalname == "library":
				found |= 2
				library = element.text
			elif elementLocalname == "name":
				found |= 4
				name = element.text
			elif elementLocalname == "version":
				found |= 8
				version = element.text
			elif elementLocalname == "description":
				found |= 16
				description = element.text
			else:
				self.Parse(element)

			if found == 31:
				break

		for element in i:
			if isinstance(element, _Comment):
				continue

			self.Parse(element)

		vlnv = VLNV(vendor=vendor, library=library, name=name, version=version)
		return vlnv, description

	@abstractmethod
	def Parse(self, element: _Element) -> None:
		pass



# ==================================================================================================================== #
#              _____ ____    _        _      ___ ______  __    _    ____ _____                                         #
#  _ __  _   _| ____|  _ \  / \      / \    |_ _|  _ \ \/ /   / \  / ___|_   _|                                        #
# | '_ \| | | |  _| | | | |/ _ \    / _ \    | || |_) \  /   / _ \| |     | |                                          #
# | |_) | |_| | |___| |_| / ___ \  / ___ \ _ | ||  __//  \  / ___ \ |___  | |                                          #
# | .__/ \__, |_____|____/_/   \_\/_/   \_(_)___|_|  /_/\_\/_/   \_\____| |_|                                          #
# |_|    |___/                                                                                                         #
# ==================================================================================================================== #
# Authors:                                                                                                             #
#   Patrick Lehmann                                                                                                    #
#                                                                                                                      #
# License:                                                                                                             #
# ==================================================================================================================== #
# Copyright 2017-2026 Patrick Lehmann - Bötzingen, Germany                                                             #
# Copyright 2016-2016 Patrick Lehmann - Dresden, Germany                                                               #
#                                                                                                                      #
# Licensed under the Apache License, Version 2.0 (the "License");                                                      #
# you may not use this file except in compliance with the License.                                                     #
# You may obtain a copy of the License at                                                                              #
#                                                                                                                      #
#   http://www.apache.org/licenses/LICENSE-2.0                                                                         #
#                                                                                                                      #
# Unless required by applicable law or agreed to in writing, software                                                  #
# distributed under the License is distributed on an "AS IS" BASIS,                                                    #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                             #
# See the License for the specific language governing permissions and                                                  #
# limitations under the License.                                                                                       #
#                                                                                                                      #
# SPDX-License-Identifier: Apache-2.0                                                                                  #
# ==================================================================================================================== #
#
"""Package installer for 'An abstract SystemRDL language model'."""
from setuptools          import setup

from itertools           import chain
from pathlib             import Path
from pyTooling.Packaging import DescribePythonPackageHostedOnGitHub, DEFAULT_CLASSIFIERS

gitHubNamespace =        "edaa-org"
packageName =            "pyEDAA.IPXACT"
packageDirectory =       packageName.replace(".", "/")
packageInformationFile = Path(f"{packageDirectory}/__init__.py")

setup(
	**DescribePythonPackageHostedOnGitHub(
		packageName=packageName,
		description="A Document-Object-Model (DOM) for IP-XACT files.",
		gitHubNamespace=gitHubNamespace,
		sourceFileWithVersion=packageInformationFile,
		classifiers=list(DEFAULT_CLASSIFIERS) + [
			"Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)"
		],
		developmentStatus="alpha",
		pythonVersions=("3.11", "3.12", "3.13", "3.14"),
		dataFiles={
			packageName: [
				str(file.relative_to(Path.cwd() / "pyEDAA/IPXACT")) for file in chain(
					Path.cwd().glob("pyEDAA/IPXACT/Schema/ipxact-*/*.xsd"),
					Path.cwd().glob("pyEDAA/IPXACT/Schema/ipxact-*/README.md"),
					Path.cwd().glob("pyEDAA/IPXACT/Schema/ieee-1685-*/*.xsd"),
					Path.cwd().glob("pyEDAA/IPXACT/Schema/ieee-1685-*/README.md"),
					Path.cwd().glob("pyEDAA/IPXACT/Schema/ieee-1685-*/LICENSE"),
					Path.cwd().glob("pyEDAA/IPXACT/Schema/ieee-1685-*/NOTICE"),
					Path.cwd().glob("pyEDAA/IPXACT/Schema/*.md"),
					Path.cwd().glob("pyEDAA/IPXACT/py.typed")
				)
			],
		}
	)
)
